/*
** Module   : Cursor.java
** Abstract : Cursor support for scrolling, dynamic queries
**
** Copyright (c) 2004-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 ECF 20051215   @23737 Created initial version. Used by dynamic
**                           queries to manage scrolling and repositioning
**                           in a progressively updated results list.
** 002 ECF 20060216   @24717 Changed beforeFirst() method access. Made package
**                           private so CompoundQuery can access it.
** 003 ECF 20061019   @30526 Added reset() method. Resets cursor to a
**                           known, default state.
** 004 ECF 20061204   @31564 Integrated RepositionCache. This replaced a
**                           simple hash map approach which failed in
**                           cases where a lookup used an ID array which
**                           had fewer IDs than joined tables in a query.
**                           Also fixed reposition(int row) method variant
**                           to ensure cursor is always positioned before
**                           the target record.
** 005 ECF 20061206   @31585 Added boolean return value to reposition()
**                           variants which accept IDs. Indicates whether
**                           desired row was matched.
** 006 ECF 20070507   @33454 Fixed forward(). Set position to end of
**                           results after attempting to advance the
**                           cache.
** 007 ECF 20070518   @33679 Added contains() method. Tests whether the
**                           cursor's cache contains a particular primary
**                           key ID.
** 008 ECF 20071011   @35444 Fixed forward() and backward() methods. It
**                           was possible for the cursor's position to get
**                           out of sync with the associated query and for
**                           duplicate, ghost records to be cached by the
**                           cursor.
** 009 ECF 20080112   @36998 Removed restriction of dealing only with IDs.
**                           This class can now handle both DMO instances
**                           and primary key IDs when retrieving data.
** 010 ECF 20080731   @39317 Added rowid support. Refactored reposition*()
**                           methods.
** 011 ECF 20081106   @40321 Fixed regression introduced by H009 (@36998).
**                           Fixed contains() method to work whether IDs
**                           or DMOs are stored by the Cursor.
** 012 ECF 20090306   @42978 Added information to be tracked. An instance now
**                           knows if it has visited the first result in a set
**                           and which direction it has run off end of query
**                           results.
** 013 SVL 20140909          Modifications that allow cursor to be used as
**                           cache for dynamic scrolling queries.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.util.*;

import org.apache.commons.logging.*;
import com.goldencode.p2j.util.*;

/**
 * Implementation of a query results list scrollable cursor, for queries which
 * fetch their results dynamically.  This object maintains a transient list
 * of results for a dynamic query, such that the user of the query may scroll
 * forward, backward, to an absolute results list row, or to a result which
 * corresponds with an array of primary key IDs (which represent a particular
 * set of joined records).  The list is considered transient because it may
 * be discarded and rebuilt in certain cases where the user changes scrolling
 * or query navigation direction.
 * <p>
 * This class is used by the {@link DynamicQuery} class and requires that the
 * query it manages implement the {@link Scrollable} interface.  It is not
 * manipulated directly by business logic.
 */
final class Cursor
{
   /** Logger */
   private static final Log LOG = Utils.getLog(Cursor.class);
   
   /** Cached list of query results (as arrays of primary keys or DMOs) */
   private final List<Object[]> results = new ArrayList<Object[]>();
   
   /** Map of primary key lists of query results to their result list rows */
   private final RepositionCache repoCache = new RepositionCache();
   
   /** Scrollable query whose results are cursored */
   private final Scrollable query;
   
   /** Current position within the results list */
   private double position = -0.5;
   
   /** Flag indicating forward scroll direction */
   private boolean forwardScroll = true;
   
   /** Have we visited the first result in the result set? */
   private boolean foundFirst = false;
   
   /** Have we visited the full result set? */
   private boolean fullSet = false;
   
   /**
    * Constructor which associates the cursor with the query it is managing.
    *
    * @param   query
    *          Instance of a scrollable query.
    */
   Cursor(Scrollable query)
   {
      this.query = query;
   }
   
   /**
    * Reset cursor to a known default state.
    */
   protected void reset()
   {
      results.clear();
      repoCache.clear();
      position = -0.5;
      foundFirst = false;
      forwardScroll = true;
      fullSet = false;
   }
   
   /**
    * Reposition the cursor such that a request to retrieve the next result
    * will retrieve the result which matches the specified primary key ID.  A
    * request to retrieve the previous result will retrieve the result
    * immediately previous to this, if any, in the results list.
    *
    * @param   id
    *          A primary key ID, representing the target record (in the case
    *          of a join, this ID is associated with the left-most record in
    *          the join).
    *
    * @return  <code>true</code> if a matching record was found and the
    *          cursor was repositioned to it;  <code>false</code> if any ID
    *          was unknown or if a matching record could not be found (and we
    *          are in silent error mode).
    * 
    * @throws  ErrorConditionException
    *          if the specified ID represents the unknown value.
    */
   boolean repositionByID(integer id)
   {
      Serializable[] serIDs = new Serializable[] { new Long(id.longValue()) };
      
      if (!validateReposition(id))
      {
         return false;
      }
         
      return reposition(serIDs, true);
   }
   
   /**
    * Reposition the cursor such that a request to retrieve the next result
    * will retrieve the result which matches the specified array of primary
    * key IDs.  A request to retrieve the previous result will retrieve the
    * result immediately previous to this, if any, in the results list.
    * <p>
    * If the specified result does not already exist in the results list, the
    * query is instructed to retrieve results sequentially, starting with the
    * first result, caching each result in its list.  This continues until
    * either the desired result is found (and the cursor is then repositioned
    * accordingly), or until no more results are available (in which case the
    * cursor remains at its current position).
    *
    * @param   id1
    *          A primary key ID, representing the target record (in the case
    *          of a join, this ID is associated with the left-most record in
    *          the join).
    * @param   joinIDs
    *          All remaining primary key IDs, if any, arranged from left to
    *          right to coincide with the records being joined by the
    *          underlying query.
    *
    * @return  <code>true</code> if a matching record was found and the
    *          cursor was repositioned to it;  <code>false</code> if any ID
    *          was unknown or if a matching record could not be found (and we
    *          are in silent error mode).
    *
    * @throws  ErrorConditionException
    *          if any of the specified IDs represents the unknown value or if
    *          we cannot reposition to the given IDs.
    */
   boolean repositionByID(rowid id1, rowid... joinIDs)
   {
      if (!validateReposition(id1))
      {
         return false;
      }
      
      int len = joinIDs.length;
      Serializable[] serIDs = new Serializable[len + 1];
      
      serIDs[0] = id1.getValue();
      
      for (int i = 0; i < len; i++)
      {
         rowid next = joinIDs[i];
         
         if (!validateReposition(next))
         {
            return false;
         }
         
         serIDs[i + 1] = next.getValue();
      }
      
      return reposition(serIDs, true);
   }
   
   /**
    * Reposition the cursor such that a request to retrieve the next result
    * will retrieve the result which matches the specified array of primary
    * key IDs.  A request to retrieve the previous result will retrieve the
    * result immediately previous to this, if any, in the results list.
    * <p>
    * If the specified result already exists in the results list, the cursor
    * is immediately repositioned to just before that result.  Otherwise,
    * the behavior depends upon the value of the <code>scan</code> parameter.
    * If it is <code>true</code>, the query is instructed to retrieve results
    * sequentially, starting with the first available, caching each result in
    * its list.  This continues until either the desired result is found (in
    * which case the cursor is repositioned accordingly), or until no more
    * results are available (in which case the cursor remains at its current
    * position).  If <code>scan</code> is <code>false</code>, this signifies
    * a logic error in the persistence library, and an IllegalStateException
    * is raised accordingly.
    *
    * @param   ids
    *          Array of primary key IDs, arranged from left to right to
    *          coincide with the records being joined by the underlying
    *          query.
    * @param   scan
    *          <code>true</code> to scan all results in the event a match is
    *          not found in the cache;  <code>false</code> to check the cache
    *          only.
    *
    * @return  <code>true</code> if a matching record was found and the
    *          cursor was repositioned to it;  <code>false</code> if any ID
    *          was unknown or if a matching record could not be found (and we
    *          are in silent error mode).
    *
    * @throws  ErrorConditionException
    *          if any of the specified IDs represents the unknown value or if
    *          we cannot reposition to the given IDs.
    * @throws  IllegalStateException
    *          if <code>scan</code> is <code>false</code> and the target
    *          result does not exist among the cached results.
    */
   boolean reposition(Serializable[] ids, boolean scan)
   {
      // Check cache first.
      int row = repoCache.getRow(ids);
      if (row >= 0)
      {
         position = forwardScroll ? row - 0.5 : row + 0.5;
         
         return true;
      }
      
      if (!scan)
      {
         return false;
      }
      
      // Scan all results for a match.
      Object[] data;

      while (true)
      {
         data = forwardScroll ? query.peekNext() : query.peekPrevious();

         if (forwardScroll)
         {
            addResultNext(data);
         }
         else
         {
            addResultPrevious(data);
         }

         if (data == null)
            break;

         if (match(ids, data))
         {
            return true;
         }
      }

      // Here we need to emulate a quirk/defect of Progress' reposition
      // implementation:  if no match was found, we leave the "current"
      // position at the last checked position.  Furthermore, if we are
      // fetching on reposition, we have to load this last record into the
      // associated buffer(s), to truly emulate the GET NEXT semantic.
      if (query.isFetchOnReposition() && !results.isEmpty())
      {
         try
         {
            query.next();
         }
         catch (ErrorConditionException exc)
         {
            if (LOG.isDebugEnabled())
            {
               LOG.debug("Error fetching record(s) during reposition", exc);
            }
         }
      }
      
      ErrorManager.recordOrThrowError(
         7331,
         "Cannot reposition query to recid/rowid(s) given");
      
      return false;
   }

   /**
    * Get off-end state of query results.
    *
    * @return off-end state of query results.
    */
   OffEnd getOffEnd()
   {
      if (!foundFirst)
         return OffEnd.NONE;

      if (fullSet && position > results.size() - 1)
         return OffEnd.BACK;

      if (position < 0)
         return forwardScroll ? OffEnd.FRONT : OffEnd.BACK;

      return OffEnd.NONE;
   }

   /**
    * Reposition the cursor to the specified row in the result set.  The row
    * indicates the position between two results, such that a call to
    * retrieve the next record will get the result at the specified row,
    * and a call to retrieve the previous result will get the result before
    * the new position.
    * <p>
    * If the specified row does not already exist in the results list, the
    * query is instructed to retrieve results sequentially, starting with the
    * first result, caching each result in its list.  This continues until
    * either the desired row is reached (and the cursor is then repositioned
    * accordingly), or until no more results are available (in which case the
    * cursor remains at its current position).
    * <p>
    * Note that this method will clear the cache if the cursor previously had
    * been scrolling backward.
    *
    * @param   row
    *          1-based index of the target position.
    *
    * @throws  ErrorConditionException
    *          if <code>row</code> represents unknown value.
    */
   void reposition(NumberType row)
   {
      if (validateReposition(row))
      {
         reposition(row.intValue());
      }
   }
   
   /**
    * Reposition the cursor to the specified row in the result set.  The row
    * indicates the position between two results, such that a call to
    * retrieve the next record will get the result at the specified row,
    * and a call to retrieve the previous result will get the result before
    * the new position.
    * <p>
    * If the specified row does not already exist in the results list, the
    * query is instructed to retrieve results sequentially, starting with the
    * first result, caching each result in its list.  This continues until
    * either the desired row is reached (and the cursor is then repositioned
    * accordingly), or until no more results are available (in which case the
    * cursor remains at its current position).
    * <p>
    * Note that this will clear the cache if we previously had been scrolling
    * backward.
    *
    * @param   row
    *          1-based index of the target position.
    */
   void reposition(int row)
   {
      // Adjust for 1-based index;  disallow negative numbers.
      row = Math.max(0, --row);
      
      // Change to forward scroll mode if necessary.
      if (!forwardScroll)
      {
         forwardScroll = true;
         results.clear();
         repoCache.clear();

         first();
         query.first();
      }
      
      // Convert to relative move.
      int jump = row - ((int) Math.ceil(position + 0.5));
      if (jump < 0)
      {
         backward(-jump);
      }
      else if (jump > 0)
      {
         forward(jump);
         getOffEnd();
      }
   }
   
   /**
    * Advance the current cursor position forward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row forward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it ahead <i>one and a half positions</i>, to just
    *       beyond the following result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it ahead <i>one position</i>, to just beyond the
    *       following result.
    * </ul>
    * <p>
    * If the request carries the cursor beyond the end of the currently
    * cached results list, the query is instructed to retrieve additional
    * results sequentially.  These are added to the results list.  This
    * continues until the request is satisfied, or until there are no more
    * results available.  In the latter case, the cursor is left positioned
    * just beyond the last available result.
    *
    * @param   rows
    *          Number of rows to scroll the cursor forward.
    *
    * @throws  ErrorConditionException
    *          if <code>rows</code> represents the unknown value.
    */
   void forward(NumberType rows)
   {
      if (validateReposition(rows))
      {
         forward(rows.intValue());
      }
   }
   
   /**
    * Advance the current cursor position forward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row forward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it ahead <i>one and a half positions</i>, to just
    *       beyond the following result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it ahead <i>one position</i>, to just beyond the
    *       following result.
    * </ul>
    * <p>
    * If the request carries the cursor beyond the end of the currently
    * cached results list, the query is instructed to retrieve additional
    * results sequentially.  These are added to the results list.  This
    * continues until the request is satisfied, or until there are no more
    * results available.  In the latter case, the cursor is left positioned
    * just beyond the last available result.
    *
    * @param   rows
    *          Number of rows to scroll the cursor forward.
    */
   void forward(int rows)
   {
      int cacheSize = results.size(); //TODO handle cacheSize = 0
      if (forwardScroll)
      {
         double gap = Math.floor(position) + 1 + rows - cacheSize;
         if (gap > 0)
         {
            if (!fullSet)
            {
               if (cacheSize > 0)
               {
                  // Advance to just before last cached result.
                  position = cacheSize - 1.5;

                  // Synchronize query's results with cursor's position.
                  query.next();
               }

               for (int i = 0; i < gap && OffEnd.NONE.equals(getOffEnd()); i++)
               {
                  Object[] row = query.peekNext();
                  addResultNext(row);
               }
            }

            position = results.size() - 1;
         }
         else
         {
            position += rows;
         }

         // Ensure cursor is between two entries.
         position = Math.floor(position) + 0.5;
      }
      else
      {
         // TODO check behavior
         position -= rows;

         // Ensure cursor is between two entries.
         position = Math.ceil(position) - 0.5;

         if (position < -0.5)
         {
            position = -0.5;
         }
      }
   }
   
   /**
    * Move the current cursor position backward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row backward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it back <i>one half position</i>, to just
    *       before the current result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it back <i>one position</i>, to just before the
    *       current result.
    * </ul>
    * <p>
    * If the request carries the cursor before the start of the currently
    * cached results list, the query is instructed to retrieve additional,
    * previous results sequentially.  These are prepended to the results list.
    * This continues until the request is satisfied, or until there are no
    * more results available.  In the latter case, the cursor is left
    * positioned just before the first available result.
    *
    * @param   rows
    *          Number of rows to scroll the cursor backward.
    *
    * @throws  ErrorConditionException
    *          if <code>rows</code> represents the unknown value.
    */
   void backward(NumberType rows)
   {
      if (validateReposition(rows))
      {
         backward(rows.intValue());
      }
   }
   
   /**
    * Move the current cursor position backward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row backward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it back <i>one half position</i>, to just
    *       before the current result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it back <i>one position</i>, to just before the
    *       current result.
    * </ul>
    * <p>
    * If the request carries the cursor before the start of the currently
    * cached results list, the query is instructed to retrieve additional,
    * previous results sequentially.  These are prepended to the results list.
    * This continues until the request is satisfied, or until there are no
    * more results available.  In the latter case, the cursor is left
    * positioned just before the first available result.
    *
    * @param   rows
    *          Number of rows to scroll the cursor backward.
    */
   void backward(int rows)
   {
      if (results.size() == 0) //TODO check behavior
      {
         return;
      }

      //if (forwardScroll) TODO separate forward and backward scroll cases
      {
         double gap = rows - position - 1;
         if (gap > 0)
         {
            if (!fullSet)
            {
               if (results.size() > 0)
               {
                  // Advance to just before first cached result.
                  position = -0.5;

                  // Synchronize query's results with cursor's position.
                  query.next();
               }

               for (int i = 0; i < gap; i++)
               {
                  Object[] row = query.peekPrevious();
                  addResultPrevious(row);

                  if (row == null)
                  {
                     break;
                  }
               }
            }

            position = -0.5;
         }
         else
         {
            position = Math.floor(position) - rows + 0.5;
         }
      }
   }
   
   /**
    * Get the <b><i>current</i></b> size of the cached results list.  However,
    * since this cache is built progressively as the query is scrolled,
    * repositioned, and naturally navigated, this <b>does not</b> necessarily
    * represent the full count of results which may satisfy the associated
    * query.
    *
    * @return  Current results list size.
    */
   integer size()
   {
      return (new integer(results.size()));
   }

   /**
    * Get the <b><i>current</i></b> size of the cached results list.  However,
    * since this cache is built progressively as the query is scrolled,
    * repositioned, and naturally navigated, this <b>does not</b> necessarily
    * represent the full count of results which may satisfy the associated
    * query.
    *
    * @return  Current results list size.
    */
   int sizeInt()
   {
      return results.size();
   }
   
   /**
    * Get the 1-based index of the current row in the result set.  This is
    * the index of the entry at or before which the cursor currently is
    * positioned.
    *
    * @return  Current row or the unknown value if the current row cannot be
    *          determined.  The current row cannot be determined if the
    *          results list is empty.
    */
   integer currentRow()
   {
      if (!foundFirst || !forwardScroll)
      {
         return (new integer());
      }
      
      // Determine row at which or before which we are positioned, adjusting
      // for 0-based index.
      int row = (int) Math.ceil(position) + 1;
      
      return (new integer(row));
   }
   
   /**
    * Indicate whether the cursor has crossed the specified boundary (front or
    * back) of the associated query's list of available results.  If
    * <code>true</code>, this indicates that the query cannot produce any more
    * results in that scroll direction.  If <code>boundary</code> is
    * <code>null</code> or <code>OffEnd.NONE</code>, assume the caller just
    * wants to know if the cursor has gone off-end at all (either direction).
    * 
    * @param   boundary
    *          One of the <code>OffEnd</code> enum types, or <code>null</code>
    *          to indicate we want to know if the cursor has gone off either
    *          end of the result list (equivalent to specifying the enum type
    *          <code>NONE</code>).
    *
    * @return  <code>true</code> if the cursor is off the specified end of its
    *          query's results list, else <code>false</code>.
    */
   boolean isOffEnd(OffEnd boundary)
   {
      OffEnd offEnd = getOffEnd();

      if (boundary == null || OffEnd.NONE.equals(boundary))
      {
         return !OffEnd.NONE.equals(offEnd);
      }
      
      return offEnd.equals(boundary);
   }
   
   /**
    * Store a result in the first position of the cached results list.  If
    * <code>ids</code> is <code>null</code>, this indicates the query has run
    * off the beginning of its available results.
    *
    * @param   data
    *          Result to add:  the primary keys of the records backing the
    *          result to be added or the records themselves (or a mix).
    */
   void addResultFirst(Object[] data)
   {
      if (data == null)
      {
         int row = (int) Math.floor(position);
         if (row == results.size())
         {
            fullSet = true;

            // numeration is now forward
            position = 0;
         }
         
         return;
      }
      
      foundFirst = true;

      // Change to forward scroll mode if necessary.
      if (!forwardScroll)
      {
         forwardScroll = true;
         results.clear();
         repoCache.clear();
      }
      
      results.add(0, data);
      first();
      updateRepoCache(data);
   }
   
   /**
    * Store a result in the last position of the cached results list.  If
    * <code>ids</code> is <code>null</code>, this indicates the query has run
    * off the end of its available results.
    *
    * @param   data
    *          Result to add:  the primary keys of the records backing the
    *          result to be added or the records themselves (or a mix).
    */
   void addResultLast(Object[] data)
   {
      if (data == null)
      {
         return;
      }

      foundFirst = true;

      // Change to backward scroll mode if necessary.
      if (forwardScroll)
      {
         forwardScroll = false;
         results.clear();
         repoCache.clear();
      }
      
      results.add(data);
      last();
      updateRepoCache(data);
   }
   
   /**
    * Add a result to the end of the cached results list.  If <code>ids</code>
    * is <code>null</code>, this indicates the query has run off the end of
    * its available results.
    *
    * @param   data
    *          Result to add:  the primary keys of the records backing the
    *          result to be added or the records themselves (or a mix).
    */
   void addResultNext(Object[] data)
   {
      if (data == null)
      {
         if (forwardScroll)
         {
            fullSet = true;
            position = results.size() - 0.5;
         }

         return;
      }
      
      if (results.isEmpty())
      {
         addResultFirst(data);
         
         return;
      }

      // Add result to end of list;  set current position to end of results
      // list.
      results.add(data);
      position = results.size() - 1;
      updateRepoCache(data);
   }

   /**
    * Prepend a result to the beginning of the cached results list.  If
    * <code>ids</code> is <code>null</code>, this indicates the query has run
    * off the beginning of its available results.
    *
    * @param   data
    *          Result to add:  the primary keys of the records backing the
    *          result to be added or the records themselves (or a mix).
    */
   void addResultPrevious(Object[] data)
   {
      if (data == null)
      {
         if (!forwardScroll)
         {
            fullSet = true;
            position = -0.5;
            forwardScroll = true;
            Collections.reverse(results);
         }

         return;
      }
      
      if (results.isEmpty())
      {
         addResultLast(data);
         
         return;
      }

      results.add(data);
      position = results.size() - 1;
      updateRepoCache(data);
   }
   
   /**
    * Move the cursor exactly onto the index of the first result <i>in the
    * cache</i>, and return the primary key array representing that result.
    * If there is no such result, the original cursor position is restored.
    * Note that such a condition does not necessarily imply that no further
    * result is available, only that none was cached by the cursor.
    *
    * @return  Primary key array representing the first cached result row, or
    *          <code>null</code> if no such result was cached.
    */
   Object[] getFirst()
   {
      if (results.isEmpty() ||
          (!forwardScroll && !fullSet))
      {
         return null;
      }

      first();

      return getResult(-0.5);
   }
   
   /**
    * Move the cursor exactly onto the index of the last result <i>in the
    * cache</i>, and return the primary key array representing that result.
    * If there is no such result, the original cursor position is restored.
    * Note that such a condition does not necessarily imply that no further
    * result is available, only that none was cached by the cursor.
    *
    * @return  Primary key array representing the last cached result row, or
    *          <code>null</code> if no such result was cached.
    */
   Object[] getLast()
   {
      if (results.isEmpty() || (!fullSet && forwardScroll))
      {
         return null;
      }
      
      last();

      return getResult(position + 0.5);
   }

   /**
    * Move the cursor exactly onto the index of the next result <i>in the
    * cache</i>, and return the primary key array representing that result.
    * If there is no such result, the original cursor position is restored.
    * Note that such a condition does not necessarily imply that no further
    * result is available, only that none was cached by the cursor.
    *
    * @return  Primary key array representing the next cached result row, or
    *          <code>null</code> if no such result was cached.
    */
   Object[] getNext()
   {
      if (results.isEmpty())
      {
         return null;
      }
      
      next();

      return getResult(forwardScroll ? position - 0.5 : position + 0.5);
   }
   
   /**
    * Move the cursor exactly onto the index of the previous result <i>in the
    * cache</i>, and return the primary key array representing that result.
    * If there is no such result, the original cursor position is restored.
    * Note that such a condition does not necessarily imply that no further
    * result is available, only that none was cached by the cursor.
    *
    * @return  Primary key array representing the previous cached result row,
    *          or <code>null</code> if no such result was cached.
    */
   Object[] getPrevious()
   {
      if (results.isEmpty())
      {
         return null;
      }
      
      previous();

      return getResult(forwardScroll ? position + 0.5 : position - 0.5);
   }
   
   /**
    * Retrieve the virtual, composite row of results at the cursor's current
    * position.  If there are no results cached, return null.
    * 
    * @return  Array of results for the current row.
    */
   Object[] getRow()
   {
      if (results.isEmpty())
      {
         return null;
      }
      
      return getResult(position);
   }
   
   /**
    * Indicate whether the results cache contains the given ID within any of
    * its arrays.  Each array within the cache is checked at the given index
    * for a match.
    *
    * @param   id
    *          ID to match.
    * @param   index
    *          Index position at which a match is attempted within each array
    *          stored in the results cache.
    *
    * @return  <code>true</code> if the cache contains the given ID, else
    *          <code>false</code>.
    */
   boolean contains(Serializable id, int index)
   {
      Iterator<Object[]> iter = results.iterator();
      while (iter.hasNext())
      {
         Object[] next = iter.next();
         Object o = next[index];
         Serializable nextID = null;
         if (o instanceof Persistable)
         {
            nextID = ((Persistable) o).getId();
         }
         else
         {
            nextID = (Serializable) o;
         }
         
         if (nextID.equals(id))
         {
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * Move the cursor to just before the index of the first entry, whether or
    * not it actually exists in the results list.
    */
   void beforeFirst()
   {
      position = -0.5;
   }

   /**
    * Determines if the query has fetched all of its results. I.e. if it has reached the last
    * result starting from the first one, of the first stating from the last one.
    *
    * @return <code>true</code> if the query has fetched all of its results.
    */
   boolean isFullSet()
   {
      return fullSet;
   }

   /**
    * Determines if the cache is build in the forward direction (starting from the first result),
    * or backward direction (starting from the last result).
    *
    * @return <code>true</code> if the cache is build in the forward direction, <code>false</code>
    *         if it is build in the backward direction
    */
   boolean isForwardScroll()
   {
      return forwardScroll;
   }

   /**
    * Determines if we have visited the first result in the result set.
    *
    * @return <code>true</code> if we have visited the first result in the result set.
    */
   boolean isFoundFirst()
   {
      return foundFirst;
   }
   
   /**
    * Move the cursor exactly onto the index of the first entry, whether or
    * not it actually exists in the results list.
    */
   private void first()
   {
      position = 0;
   }
   
   /**
    * Move the cursor exactly onto the index of the last entry, whether or
    * not it actually exists in the results list.
    */
   private void last()
   {
      if (forwardScroll)
      {
         position = Math.max(0, results.size() - 1);
      }
      else
      {
         position = 0;
      }
   }
   
   /**
    * Move the cursor exactly onto the index of the next entry, whether or
    * not it actually exists in the results list.
    */
   public void next()
   {
      if (forwardScroll)
      {
         position = Math.floor(position) + 1;
      }
      else
      {
         position = Math.ceil(position) - 1;
      }
   }
   
   /**
    * Move the cursor exactly onto the index of the previous entry, whether
    * or not it actually exists in the results list.
    */
   private void previous()
   {
      if (forwardScroll)
      {
         position = Math.ceil(position) - 1;
      }
      else
      {
         position = Math.floor(position) + 1;
      }
   }
   
   /**
    * Retrieve the result at the cursor's current position, which should
    * always be at an index boundary (not between result indices).  If there
    * is no such result, the cursor's current position is reset to the
    * position provided and <code>null</code> is returned.  The latter
    * condition indicates that the cursor has been moved beyond a cached
    * result, such that the caller needs to attempt to query a result and
    * cache it.
    *
    * @param   resetPosition
    *          Position to which the cursor must be reset in the event no
    *          result is available in the cached results list at the cursor's
    *          current position.
    * 
    * @return  The data at the cursor's current position, or <code>null</code>
    *          if no such data exists.
    */
   private Object[] getResult(double resetPosition)
   {
      // It is safe to truncate here, since position is expected to be set to
      // an integer boundary when this method is called.
      int index = (int) position;
      
      if (index < 0 || index >= results.size())
      {
         position = resetPosition;
         return null;
      }
      
      return results.get(index);
   }
   
   /**
    * Associate an array of primary key IDs representing a query result with
    * the cursor's current position.  This is used to enable a fast look up
    * when performing a "reverse" reposition;  that is, a reposition where
    * the cursor is given an array of primary keys and must move to the
    * associated results list row.
    * <p>
    * Note that this method should only be invoked when the current cursor
    * position corresponds with the specified result, otherwise the row will
    * be mismatched for the result.
    *
    * @param   data
    *          Array of primary keys or DMOs representing a query result.
    */
   private void updateRepoCache(Object[] data)
   {
      int row = (int) Math.ceil(position);
      int len = data.length;
      Serializable[] ids = new Serializable[len];
      for (int i = 0; i < len; i++)
      {
         Object datum = data[i];
         ids[i] = (datum instanceof Persistable
                   ? ((Persistable) datum).getId()
                   : (Serializable) datum);
      }
      repoCache.add(ids, row);
   }
   
   /**
    * Detect whether all elements in one object array match those at the same
    * positions within another array.  The length of the first array
    * determines the number of elements checked.  That is, if the second
    * array contains more elements than the first, excess elements in the
    * second array are ignored.
    *
    * @param   a1
    *          First array.  May have fewer or the same number of elements as
    *          <code>a2</code>.  Elements may not be <code>null</code>.
    * @param   a2
    *          Second array.  May have the same number or more elements than
    *          <code>a1</code>.  Elements may not be <code>null</code>.
    *
    * @return  <code>true</code> if all elements in <code>a1</code> equal
    *          those elements at corresponding positions in <code>a2</code>.
    */
   private boolean match(Serializable[] a1, Object[] a2)
   {
      int len = a1.length;
      for (int i = 0; i < len; i++)
      {
         Serializable id = a1[i];
         Object datum = a2[i];
         
         if (datum instanceof Persistable)
         {
            datum = ((Persistable) datum).getId();
         }
         
         if (!id.equals(datum))
         {
            return false;
         }
      }
      
      return true;
   }
   
   /**
    * Validate a numeric value provided to a reposition request to ensure
    * that it does not represent the unknown value.
    *
    * @param   value
    *          ID or row number/offset to be validated.
    *
    * @return  <code>true</code> if the number is valid;  <code>false</code>
    *          if it is invalid, but we are in silent error mode.
    *
    * @throws  ErrorConditionException
    *          if the number is invalid and we are not in silent error mode.
    */
   private boolean validateReposition(BaseDataType value)
   {
      if (value.isUnknown())
      {
         ErrorManager.recordOrThrowError(
            3165,
            "Could not evaluate reposition amount for query");
         
         return false;
      }
      
      return true;
   }
}
